@@ -27,6 +27,8 @@ import traceback
 
 import docker
 
+from distutils.version import StrictVersion
+
 from ansible.module_utils.basic import AnsibleModule
 
 DOCUMENTATION = '''
@@ -149,6 +151,16 @@ options:
     default: None
     choices:
       - host
+  cgroupns_mode:
+    description:
+      - Set docker cgroups namespace (default depends on Docker config)
+      - Supported only with Docker 20.10 (Docker API 1.41) and later
+    required: False
+    type: str
+    default: None
+    choices:
+      - private
+      - host
   privileged:
     description:
       - Set the container to privileged
@@ -176,6 +188,11 @@ options:
       - How many times to attempt a restart if 'on-failure' policy is set
     type: int
     default: 10
+  tmpfs:
+    description:
+      - List of paths to mount as tmpfs.
+    required: False
+    type: list
   volumes:
     description:
       - Set volumes for docker to use
@@ -247,9 +264,6 @@ EXAMPLES = '''
 '''
 
 
-COMPARE_CONFIG_CMD = ['/usr/local/bin/kolla_set_configs', '--check']
-
-
 def get_docker_client():
     return docker.APIClient
 
@@ -273,6 +287,9 @@ class DockerWorker(object):
 
         self.dc = get_docker_client()(**options)
 
+        self._cgroupns_mode_supported = (
+            StrictVersion(self.dc._version) >= StrictVersion('1.41'))
+
     def generate_tls(self):
         tls = {'verify': self.params.get('tls_verify')}
         tls_cert = self.params.get('tls_cert'),
@@ -331,9 +348,7 @@ class DockerWorker(object):
 
     def compare_container(self):
         container = self.check_container()
-        if (not container or
-                self.check_container_differs() or
-                self.compare_config()):
+        if not container or self.check_container_differs():
             self.changed = True
         return self.changed
 
@@ -347,6 +362,8 @@ class DockerWorker(object):
             self.compare_labels(container_info) or
             self.compare_privileged(container_info) or
             self.compare_pid_mode(container_info) or
+            self.compare_cgroupns_mode(container_info) or
+            self.compare_tmpfs(container_info) or
             self.compare_volumes(container_info) or
             self.compare_volumes_from(container_info) or
             self.compare_environment(container_info) or
@@ -401,6 +418,21 @@ class DockerWorker(object):
         if new_pid_mode != current_pid_mode:
             return True
 
+    def compare_cgroupns_mode(self, container_info):
+        if not self._cgroupns_mode_supported:
+            return False
+        new_cgroupns_mode = self.params.get('cgroupns_mode')
+        if new_cgroupns_mode is None:
+            # means we don't care what it is
+            return False
+        current_cgroupns_mode = (container_info['HostConfig']
+                                 .get('CgroupnsMode'))
+        if current_cgroupns_mode == '':
+            # means the container was created on Docker pre-20.10
+            # it behaves like 'host'
+            current_cgroupns_mode = 'host'
+        return new_cgroupns_mode != current_cgroupns_mode
+
     def compare_privileged(self, container_info):
         new_privileged = self.params.get('privileged')
         current_privileged = container_info['HostConfig']['Privileged']
@@ -438,6 +470,17 @@ class DockerWorker(object):
         if new_labels != current_labels:
             return True
 
+    def compare_tmpfs(self, container_info):
+        new_tmpfs = self.generate_tmpfs()
+        current_tmpfs = container_info['HostConfig'].get('Tmpfs')
+        if not new_tmpfs:
+            new_tmpfs = []
+        if not current_tmpfs:
+            current_tmpfs = []
+
+        if set(current_tmpfs).symmetric_difference(set(new_tmpfs)):
+            return True
+
     def compare_volumes_from(self, container_info):
         new_vols_from = self.params.get('volumes_from')
         current_vols_from = container_info['HostConfig'].get('VolumesFrom')
@@ -571,43 +614,6 @@ class DockerWorker(object):
             if current_healthcheck:
                 return True
 
-    def compare_config(self):
-        try:
-            job = self.dc.exec_create(
-                self.params['name'],
-                COMPARE_CONFIG_CMD,
-                user='root',
-            )
-            output = self.dc.exec_start(job)
-            exec_inspect = self.dc.exec_inspect(job)
-        except docker.errors.APIError as e:
-            # NOTE(yoctozepto): If we have a client error, then the container
-            # cannot be used for config check (e.g., is restarting, or stopped
-            # in the mean time) - assume config is stale = return True.
-            # Else, propagate the server error back.
-            if e.is_client_error():
-                return True
-            else:
-                raise
-        # Exit codes:
-        # 0: not changed
-        # 1: changed
-        # 137: abrupt exit -> changed
-        # else: error
-        if exec_inspect['ExitCode'] == 0:
-            return False
-        elif exec_inspect['ExitCode'] == 1:
-            return True
-        elif exec_inspect['ExitCode'] == 137:
-            # NOTE(yoctozepto): This is Docker's command exit due to container
-            # exit. It means the container is unstable so we are better off
-            # marking it as requiring a restart due to config update.
-            return True
-        else:
-            raise Exception('Failed to compare container configuration: '
-                            'ExitCode: %s Message: %s' %
-                            (exec_inspect['ExitCode'], output))
-
     def parse_image(self):
         full_image = self.params.get('image')
 
@@ -679,6 +685,13 @@ class DockerWorker(object):
                 if self.check_container():
                     raise
 
+    def generate_tmpfs(self):
+        tmpfs = self.params.get('tmpfs')
+        if tmpfs:
+            # NOTE(mgoddard): Filter out any empty strings.
+            tmpfs = [t for t in tmpfs if t]
+        return tmpfs
+
     def generate_volumes(self):
         volumes = self.params.get('volumes')
         if not volumes:
@@ -753,6 +766,7 @@ class DockerWorker(object):
             'security_opt': self.params.get('security_opt'),
             'pid_mode': self.params.get('pid_mode'),
             'privileged': self.params.get('privileged'),
+            'tmpfs': self.generate_tmpfs(),
             'volumes_from': self.params.get('volumes_from')
         }
 
@@ -777,7 +791,16 @@ class DockerWorker(object):
         if binds:
             options['binds'] = binds
 
-        return self.dc.create_host_config(**options)
+        host_config = self.dc.create_host_config(**options)
+
+        if self._cgroupns_mode_supported:
+            # NOTE(yoctozepto): python-docker does not support CgroupnsMode
+            # natively so we stuff it in manually.
+            cgroupns_mode = self.params.get('cgroupns_mode')
+            if cgroupns_mode is not None:
+                host_config['CgroupnsMode'] = cgroupns_mode
+
+        return host_config
 
     def _inject_env_var(self, environment_info):
         newenv = {
@@ -1091,6 +1114,8 @@ def generate_module():
         cap_add=dict(required=False, type='list', default=list()),
         security_opt=dict(required=False, type='list', default=list()),
         pid_mode=dict(required=False, type='str', choices=['host', '']),
+        cgroupns_mode=dict(required=False, type='str',
+                           choices=['private', 'host']),
         privileged=dict(required=False, type='bool', default=False),
         graceful_timeout=dict(required=False, type='int', default=10),
         remove_on_exit=dict(required=False, type='bool', default=True),
@@ -1108,6 +1133,7 @@ def generate_module():
         tls_cert=dict(required=False, type='str'),
         tls_key=dict(required=False, type='str'),
         tls_cacert=dict(required=False, type='str'),
+        tmpfs=dict(required=False, type='list'),
         volumes=dict(required=False, type='list'),
         volumes_from=dict(required=False, type='list'),
         dimensions=dict(required=False, type='dict', default=dict()),
